@@ -2,19 +2,19 @@ use strict;
 use warnings;
 
 package Test::Deep::PDL;
-$Test::Deep::PDL::VERSION = '0.20';
+$Test::Deep::PDL::VERSION = '0.21';
 # ABSTRACT: Test ndarrays inside data structures with Test::Deep
 
 
 use Test::Deep::Cmp;
-use Test::PDL qw( eq_pdl_diag );
+require Test::PDL;
 
 
 sub init
 {
 	my $self = shift;
 	my $expected = shift;
-	die "Supplied value is not a ndarray" unless eval { $expected->isa('PDL') };
+	die "Supplied value is not an ndarray" unless eval { $expected->isa('PDL') };
 	$self->{expected} = $expected;
 }
 
@@ -22,8 +22,7 @@ sub descend
 {
 	my $self = shift;
 	my $got = shift;
-	my( $ok, $diag ) = Test::PDL::eq_pdl_diag( $got, $self->{expected} );
-	$self->data->{diag} = $diag;
+	( my $ok, $self->data->{diag} ) = Test::PDL::eq_pdl( $got, $self->{expected} );
 	return $ok;
 }
 
@@ -31,7 +30,7 @@ sub diag_message
 {
 	my $self = shift;
 	my $where = shift;
-	return "Comparing $where as a ndarray:\n" . $self->data->{diag};
+	return "Comparing $where as an ndarray:\n" . $self->data->{diag};
 }
 
 sub renderExp
@@ -63,7 +62,7 @@ Test::Deep::PDL - Test ndarrays inside data structures with Test::Deep
 
 =head1 VERSION
 
-version 0.20
+version 0.21
 
 =head1 DESCRIPTION
 
@@ -87,7 +86,7 @@ Edward Baudrez <ebaudrez@cpan.org>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2022 by Edward Baudrez.
+This software is copyright (c) 2024 by Edward Baudrez.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
