@@ -1,5 +1,5 @@
 package Test::PDL;
-$Test::PDL::VERSION = '0.20';
+$Test::PDL::VERSION = '0.21';
 # ABSTRACT: Test Perl Data Language arrays (a.k.a. ndarrays) for equality
 
 
@@ -7,16 +7,18 @@ use strict;
 use warnings;
 use PDL::Lite;
 use PDL::Types ();
+use Test::PDL::Helper 'approx_artol';
 
 use base qw( Exporter );
 our @EXPORT = qw( is_pdl );
-our @EXPORT_OK = qw( eq_pdl eq_pdl_diag is_pdl test_pdl );
+our @EXPORT_OK = qw( eq_pdl is_pdl test_pdl );
 our %EXPORT_TAGS = ( deep => [ qw( test_pdl ) ] );
 
 
-our %OPTIONS = (
-	TOLERANCE   => 1e-6,
-	EQUAL_TYPES => 1,
+our %DEFAULTS = (
+	atol                => 1e-6,
+	require_equal_types => 1,
+	rtol                => 1e-6,
 );
 
 
@@ -26,8 +28,10 @@ sub import
 	while( $i < @_ ) {
 		if( $_[ $i ] =~ /^-/ ) {
 			my( $key, $val ) = splice @_, $i, 2;
-			$key =~ s/^-(.*)/\U$1/;
-			set_options( $key, $val );
+			$key =~ s/^-(.*)/$1/;
+			PDL::barf( "invalid name $key" ) unless grep { $key eq $_ } keys %DEFAULTS;
+			PDL::barf( "undefined value for $key" ) unless defined $val;
+			$DEFAULTS{ $key } = $val;
 		}
 		else { $i++ }
 	}
@@ -35,107 +39,73 @@ sub import
 }
 
 
-sub _approx
-{
-	my( $a, $b ) = @_;
-	return abs( $a - $b ) < $OPTIONS{ TOLERANCE };
+sub is_pdl {
+  require Test::Builder;
+  my ( $got, $expected, $arg ) = @_;
+  my $tb = Test::Builder->new;
+  $tb->croak('error in arguments: third argument is an ndarray')
+    if eval { $arg->isa('PDL') };
+  my $opt = { %DEFAULTS };
+  my $name;
+  if ($arg) {
+    if (ref $arg eq 'HASH') { $opt = { %$opt, %$arg } }
+    else { $name = $arg }
+  }
+  $name ||= $opt->{test_name} || "ndarrays are equal";
+  my ($ok, $reason, $mask) = eq_pdl($got, $expected, $opt);
+  return $tb->ok(1, $name) if $ok;
+  my $rc = $tb->ok( 0, $name );
+  my $fmt = '%-8T %-12D (%-5S) ';
+  my @mismatch;
+  if (defined $mask) {
+    my $coords = defined $mask ? $mask->not->whichND : undef;
+    $coords = $coords->slice(',0:4') if defined $coords and $coords->dim(1) > 5;
+    push @mismatch, (
+      "\nFirst <=5 values differ at:", $coords,
+      "Those 'got' values: ", $got->indexND($coords),
+      "\nThose 'expected' values: ", $expected->indexND($coords),
+    );
+  }
+  $tb->diag(
+    "    $reason\n",
+    "         got: ", eval { $got->isa('PDL')      && !$got->isnull      } ? $got->info( $fmt )      : '', $got, "\n",
+    "    expected: ", eval { $expected->isa('PDL') && !$expected->isnull } ? $expected->info( $fmt ) : '', $expected,
+    @mismatch,
+  );
+  return $rc;
 }
 
 
-sub _comparison_fails
-{
-	my ( $got, $expected ) = @_;
-	if( not eval { $got->isa('PDL') } ) {
-		return 'received value is not a ndarray';
-	}
-	if( not eval { $expected->isa('PDL') } ) {
-		return 'expected value is not a ndarray';
-	}
-	if( $OPTIONS{ EQUAL_TYPES } && $got->type != $expected->type ) {
-		return 'types do not match (EQUAL_TYPES is true)';
-	}
-	if( $got->ndims != $expected->ndims ) {
-		return 'dimensions do not match in number';
-	}
-	if( not _dimensions_match( [$got->dims], [$expected->dims] ) ) {
-		return 'dimensions do not match in extent';
-	}
-	# evaluating these only makes sense for ndarrays that conform in shape
-	if( ( $got->badflag == 1 || $expected->badflag == 1 ) &&
-		not eval { PDL::all( PDL::isbad($got) == PDL::isbad($expected) ) } ) {
-		return 'bad value patterns do not match';
-	}
-	return 'values do not match'
-		if ( $got->isempty and !$expected->isempty)
-		or (!$got->isempty and  $expected->isempty);
-	return 0 if $got->isempty and $expected->isempty;
-	# if we get here, bad value patterns are sure to match, remove
-	my $isgood = $got->isgood;
-	$got = $got->where($isgood), $expected = $expected->where($isgood);
-	return 0 if $got->isempty;
-	if( $got->type < PDL::float && $expected->type < PDL::float ) {
-		if( not eval { PDL::all( $got == $expected ) } ) {
-			return 'values do not match';
-		}
-	}
-	else {
-		# floating-point comparison must be approximate
-		if( not eval { PDL::all( _approx $got, $expected ) } ) {
-			return 'values do not match';
-		}
-	}
-	# if we get here, we didn't fail
-	return 0;
-}
-
-
-sub _dimensions_match
-{
-	my @A = @{ +shift };
-	my @B = @{ +shift };
-	while( my $a = shift @A and my $b = shift @B ) {
-		if( $a != $b ) { return 0 }
-	}
-	return 1;
-}
-
-
-sub is_pdl
-{
-	require Test::Builder;
-	my ( $got, $expected, $name ) = @_;
-	my $tb = Test::Builder->new;
-	if( eval { $name->isa('PDL') } ) {
-		$tb->croak( 'error in arguments: test name is a ndarray' );
-	}
-	$name ||= "ndarrays are equal";
-	if( my $reason = _comparison_fails $got, $expected ) {
-		my $rc = $tb->ok( 0, $name );
-		my $fmt = '%-8T %-12D (%-5S) ';
-		$tb->diag( "    $reason\n",
-			   "         got: ", eval { $got->isa('PDL')      && !$got->isnull      } ? $got->info( $fmt )      : '', $got, "\n",
-			   "    expected: ", eval { $expected->isa('PDL') && !$expected->isnull } ? $expected->info( $fmt ) : '', $expected );
-		return $rc;
-	}
-	else {
-		return $tb->ok( 1, $name );
-	}
-}
-
-
-sub eq_pdl
-{
-	my ( $got, $expected ) = @_;
-	return !_comparison_fails( $got, $expected );
-}
-
-
-sub eq_pdl_diag
-{
-	my ( $got, $expected ) = @_;
-	my $reason = _comparison_fails( $got, $expected );
-	if( $reason ) { return 0, $reason }
-	else { return 1 }
+sub eq_pdl {
+  my ($got, $expected, $arg) = @_;
+  my $opt = { %DEFAULTS, ref $arg eq 'HASH' ? %$arg : () };
+  PDL::barf( 'need an absolute or a relative tolerance, or both' ) unless defined $opt->{atol} || defined $opt->{rtol};
+  $opt->{atol} //= 0;
+  $opt->{rtol} //= 0;
+  PDL::barf('absolute tolerance cannot be negative') if $opt->{atol} < 0;
+  PDL::barf('relative tolerance cannot be negative') if $opt->{rtol} < 0;
+  return wantarray ? (0, 'received value is not an ndarray', undef) : 0
+    if !eval { $got->isa('PDL') };
+  return wantarray ? (0, 'expected value is not an ndarray', undef) : 0
+    if !eval { $expected->isa('PDL') };
+  return wantarray ? (0, 'types do not match (\'require_equal_types\' is true)', undef) : 0
+    if $opt->{require_equal_types} && $got->type != $expected->type;
+  my @got_dims = $got->dims;
+  my @exp_dims = $expected->dims;
+  return wantarray ? (0, 'dimensions do not match in number', undef) : 0
+    if @got_dims != @exp_dims;
+  while (@got_dims) {
+    return wantarray ? (0, 'dimensions do not match in extent', undef) : 0
+      if shift(@got_dims) != shift(@exp_dims);
+  }
+  return wantarray ? (1, '', undef) : 1
+    if $got->isempty and $expected->isempty;
+  # both are now non-empty
+  my $res = approx_artol( $got, $expected, @$opt{qw(atol rtol)} );
+  return wantarray ? (1, '', undef) : 1 if $res->all;
+  my $exp_nelem = $expected->nelem;
+  my $reason = ($exp_nelem-$res->sum)."/$exp_nelem values do not match";
+  return wantarray ? (0, $reason, $res) : 0;
 }
 
 
@@ -166,16 +136,6 @@ for my $type ( PDL::Types::types ) {
 }
 
 
-sub set_options
-{
-	while( my( $key, $value ) = splice @_, 0, 2 ) {
-		PDL::barf( "invalid option $key" ) unless grep { $key eq $_ } keys %OPTIONS;
-		PDL::barf( "undefined value for $key" ) unless defined $value;
-		$OPTIONS{ $key } = $value;
-	}
-}
-
-
 1;
 
 __END__
@@ -190,7 +150,7 @@ Test::PDL - Test Perl Data Language arrays (a.k.a. ndarrays) for equality
 
 =head1 VERSION
 
-version 0.20
+version 0.21
 
 =head1 SYNOPSIS
 
@@ -214,10 +174,17 @@ version 0.20
 	# not ok 2 - demonstrate the output of a failing test
 	#
 	#   Failed test 'demonstrate the output of a failing test'
-	#   at aux/pod.t line 16.
-	#     values do not match
+	#   at maint/pod.t line 16.
+	#     2/5 values do not match
 	#          got: Double   D [5]        (P    ) [0 -1 -2 3 4]
 	#     expected: Double   D [5]        (P    ) [0 1 2 3 4]
+	# First <=5 values differ at:
+	# [
+	#  [1]
+	#  [2]
+	# ]
+	# Those 'got' values: [-1 -2]
+	# Those 'expected' values: [1 2]
 
 	# ndarrays within other data structures can be tested with Test::Deep
 	use Test::Deep qw( cmp_deeply );
@@ -228,8 +195,8 @@ version 0.20
 	# not ok 3 - demonstrate the output of a failing deep comparison
 	#
 	#   Failed test 'demonstrate the output of a failing deep comparison'
-	#   at aux/pod.t line 30.
-	# Comparing $data->{"data"} as a ndarray:
+	#   at maint/pod.t line 30.
+	# Comparing $data->{"data"} as an ndarray:
 	# dimensions do not match in extent
 	#    got : Long     D [3]        (P    ) [17 0 1]
 	# expect : Long     D [4]        (P    ) [17 0 0 1]
@@ -239,7 +206,8 @@ version 0.20
 With Test::PDL, you can compare two ndarrays for equality. The comparison is
 performed as thoroughly as possible, comparing types, dimensions, bad value
 patterns, and finally the values themselves. The exact behaviour can be
-configured by setting certain options (see set_options() and %OPTIONS below).
+configured by setting certain package-wide defaults (see %DEFAULTS below), or
+by supplying options in a function call.
 Test::PDL is mostly useful in test scripts.
 
 Test::PDL is to be used with the Perl Data Language (L<PDL>).
@@ -251,22 +219,19 @@ with C<test_>: test_short(), test_double(), ...
 
 =head1 VARIABLES
 
-=head2 %OPTIONS
+=head2 %DEFAULTS
 
-The comparison criteria used by Test::PDL can be configured by setting the
-values in the %OPTIONS hash. This can be done directly, by addressing
-%Test::PDL::OPTIONS directly. However, it is preferred that set_options() is
-used instead.
+The default comparison criteria used by Test::PDL can be configured by setting
+the values in the %DEFAULTS hash. This can be done directly, by addressing
+%Test::PDL::DEFAULTS directly.
 
 =over 4
 
-=item TOLERANCE
+=item atol
 
-The tolerance used to compare floating-point values. Initially set to 1e-6.
-This is currently an absolute tolerance, meaning that two values compare equal
-if the absolute value of their difference is below the tolerance.
+The absolute tolerance used to compare values. Initially set to 1e-6.
 
-=item EQUAL_TYPES
+=item require_equal_types
 
 If true, only ndarrays with equal type can be considered equal. If false, the
 types of the ndarrays being compared is not taken into consideration. Defaults
@@ -276,39 +241,76 @@ write tests like
 	is_pdl( $got, pdl([ 1, 3, 5, 6 ]) );
 
 without having to worry about the type of the ndarray being exactly I<double>
-(which is the default type of the pdl() constructor), set EQUAL_TYPES equal to
+(which is the default type of the pdl() constructor), set I<require_equal_types> equal to
 0.
 
+=item rtol
+
+The relative tolerance used to compare values. Initially set to 1e-6.
+
 =back
 
 =head1 FUNCTIONS
 
 =head2 import
 
-Custom importer that recognizes configuration options specified at use time, as
+Custom importer that recognizes configuration defaults specified at use time, as
 in
 
-	use Test::PDL -equal_types => 0;
+	use Test::PDL -require_equal_types => 0;
 
-This invocation is equivalent to
+=head2 is_pdl
 
-	use Test::PDL;
-	Test::PDL::set_options( EQUAL_TYPES => 0 );
+=for ref # PDL
 
-but is arguably somewhat nicer.
+Run a test comparing an ndarray to an expected ndarray, and fail with detailed
+diagnostics if they don't compare equal.
 
-=head2 _approx
+=for usage # PDL
 
-Internal function reimplementing the functionality of PDL::approx(), but with a
-tolerance that is not remembered across invocations. Rather, the tolerance can
-be set by the user (see set_options() and $OPTIONS{TOLERANCE}), and defaults to
-1e-6.
+	is_pdl( $got, $expected );
+	is_pdl( $got, $expected, $test_name );
+	is_pdl( $got, $expected, { test_name => $test_name } );
+	is_pdl( $got, $expected, { atol => $absolute_tolerance, ... } );
 
-=head2 _comparison_fails
+Yields ok if the first two arguments are ndarrays that compare equal, not ok if
+the ndarrays are different, or if at least one is not an ndarray. Prints a
+diagnostic when the comparison fails, with the reason and a brief printout of
+both arguments. See the documentation of eq_pdl() for the comparison
+criteria. $test_name is optional.
 
-Internal function which does the real work of comparing two ndarrays. If the
-comparison fails, _comparison_fails() returns a string containing the reason for
-failure. If the comparison succeeds, _comparison_fails() returns zero.
+Named after is() from L<Test::More>.
+
+=head2 eq_pdl
+
+=for ref # PDL
+
+Return true if two ndarrays compare equal, false otherwise. In list context,
+additionally returns a diagnostic string.
+
+=for usage # PDL
+
+	my $equal = eq_pdl( $got, $expected );
+	my $equal = eq_pdl( $got, $expected, { atol => $absolute_tolerance, ... } );
+	my( $equal, $diag ) = eq_pdl( $got, $expected );
+	my( $equal, $diag ) = eq_pdl( $got, $expected, { atol => $absolute_tolerance, ... } );
+
+eq_pdl() contains just the comparison part of is_pdl(), without the
+infrastructure required to write tests with L<Test::More>. It could be used as
+part of a larger test in which the equality of two ndarrays must be verified. By
+itself, eq_pdl() does not generate any output, so it should be safe to use
+outside test suites.
+
+In list context, eq_pdl() returns a list with three elements, the first one being
+a boolean whether the ndarrays compared equal, the second being a diagnostic
+string explaining why the comparison failed (or the empty string, if it didn't
+fail). The third is either the mask of not-equal if the values didn't
+match, or C<undef>.
+This is useful in combination with L<Test::Deep>, but might also be
+useful on its own.
+
+eq_pd() does not need L<Test::Builder>, so you can use it as part of something
+else, without side effects (like generating output).
 
 The criteria for equality are the following:
 
@@ -321,7 +323,7 @@ is no implicit conversion from scalar to ndarray.
 
 =item *
 
-The type of both ndarrays must be equal if (and only if) EQUAL_TYPES is true.
+The type of both ndarrays must be equal if (and only if) I<require_equal_types> is true.
 
 =item *
 
@@ -331,9 +333,9 @@ compares equal with another two-dimensional ndarray.
 =item *
 
 The extent of the dimensions are compared one by one and must match. That is, a
-ndarray with dimensions (5,4) cannot compare equal with a ndarray of dimensions
+ndarray with dimensions (5,4) cannot compare equal with an ndarray of dimensions
 (5,3). Note that degenerate dimensions are not treated specially, and thus a
-ndarray with dimensions (5,4,1) is considered different from a ndarray with
+ndarray with dimensions (5,4,1) is considered different from an ndarray with
 dimensions (5,4).
 
 =item *
@@ -345,77 +347,24 @@ bad flag is different, if there are no bad values.
 
 =item *
 
-And last but not least, the values themselves are examined one by one. For
-integer types, the comparison is performed exactly, whereas an approximate
-equality is used for floating-point types. The approximate comparison is
-implemented using a private reimplementation of PDL::approx(). See _approx()
-for more information.
+And last but not least, the values themselves are examined one by one.
+As of 0.21, both integer and floating-point types are compared approximately.
+The approximate comparison is
+implemented using a combination of relative and absolute tolerances, which can
+be set by supplying an argument to C<use Test::PDL>, or by supplying an
+optional hash to this function. By default, the absolute and relative
+tolerances are both equal to 1e-6. The user can specify a pure relative
+tolerance by specifying C<atol =E<gt> 0>, and a pure absolute tolerance by
+specifying C<rtol =E<gt> 0>. If both tolerances are specified, values compare
+equal if I<either> their difference is lower than or equal to the absolute
+tolerance I<or> their relative difference (with respect to the expected
+value) is lower than or equal to the relative tolerance. For expected
+values equal to zero, relative differences (with respect to the expected
+value) make no sense, and the use of combined absolute and relative
+tolerances is recommended.
 
 =back
 
-=head2 _dimensions_match
-
-Internal function which compares the extent of each of the dimensions of two
-ndarrays, one by one. The dimensions must be passed in as two array references.
-Returns 1 if all dimensions match pairwise. Returns 0 otherwise.
-
-This function will not operate correctly if the number of dimensions does not
-match between the ndarrays, so be sure to check that before calling this
-function.
-
-=head2 is_pdl
-
-=for ref # PDL
-
-Run a test comparing a ndarray to an expected ndarray, and fail with detailed
-diagnostics if they don't compare equal.
-
-=for usage # PDL
-
-	is_pdl( $got, $expected, $test_name );
-
-Yields ok if the first two arguments are ndarrays that compare equal, not ok if
-the ndarrays are different, or if at least one is not a ndarray. Prints a
-diagnostic when the comparison fails, with the reason and a brief printout of
-both arguments. See the documentation of _comparison_fails() for the comparison
-criteria. $test_name is optional.
-
-Named after is() from L<Test::More>.
-
-=head2 eq_pdl
-
-=for ref # PDL
-
-Return true if two ndarrays compare equal, false otherwise.
-
-=for usage # PDL
-
-	my $equal = eq_pdl( $got, $expected );
-
-eq_pdl() contains just the comparison part of is_pdl(), without the
-infrastructure required to write tests with Test::More. It could be used as
-part of a larger test in which the equality of two ndarrays must be verified. By
-itself, eq_pdl() does not generate any output, so it should be safe to use
-outside test suites.
-
-=head2 eq_pdl_diag
-
-=for ref # PDL
-
-Return true if two ndarrays compare equal, false otherwise, and the reason why
-the comparison failed (if it did).
-
-=for usage # PDL
-
-	my( $ok ) = eq_pdl_diag( $got, $expected );
-	my( $ok, $diag ) = eq_pdl_diag( $got, $expected );
-
-eq_pdl_diag() is like eq_pdl(), except that it also returns the reason why the
-comparison failed (if it failed). $diag will be false if the comparison
-succeeds. Does not need L<Test::Builder>, so you can use it as part of
-something else, without side effects (like generating output). It was written
-to support deep comparisons with L<Test::Deep>.
-
 =head2 test_pdl
 
 =for ref # PDL
@@ -461,13 +410,13 @@ same thing with
 
 	my $expected = {
 		...,
-		some_field => code( sub { eq_pdl_diag( shift, pdl( ... ) ) } ),
+		some_field => code( sub { eq_pdl( shift, pdl( ... ) ) } ),
 		...
 	};
 
 but the diagnostics provided by test_pdl() are better, and it's easier to use.
 test_pdl() accepts the same arguments as the PDL constructor pdl() does. If you
-need to compare a ndarray with a type different from the default type, use one
+need to compare an ndarray with a type different from the default type, use one
 of the provided test_byte(), test_short(), test_long(), etc.:
 
 	my $expected = { data => test_short( -4,-9,13 ) };
@@ -489,26 +438,6 @@ above example.
 test_indx test_longlong test_float test_double test_cfloat test_cdouble
 test_cldouble test_ldouble test_sbyte test_ulong test_ulonglong
 
-=head2 set_options
-
-=for ref # PDL
-
-Configure the comparison carried out by Test::PDL's testing functions.
-
-=for example # PDL
-
-	# e.g., if a tolerance of 1e-6 is too tight
-	Test::PDL::set_options( TOLERANCE => 1e-4 );
-
-The preferred way to set the options to this module. See %OPTIONS for all
-allowed options. set_options() dies with an error if an unknown option is
-passed. Note that sensible default values are provided for all options, so you
-needn't use this routine if you are fine with the defaults.
-
-This function is not exported. Rather, it must be called as
-
-	Test::PDL::set_options( KEY => VALUE, ... );
-
 =head1 BUGS
 
 None reported so far.
@@ -522,13 +451,16 @@ L<PDL>, L<Test::More>, L<Test::Deep>, L<Test::PDL::Deep>
 Thanks to PDL Porters Joel Berger, Chris Marshall, and David Mertens for
 feedback and improvements.
 
+Thanks to Ed J, Zakariyya Mughal, and Diab Jerius for feedback, improvements,
+maintenance of the code, and encouragement!
+
 =head1 AUTHOR
 
 Edward Baudrez <ebaudrez@cpan.org>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2022 by Edward Baudrez.
+This software is copyright (c) 2024 by Edward Baudrez.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
